from work_stream.block import StreamBlock


class StreamLine:
    """
        流水线，用于编排块。
    """

    def __init__(self):
        """
            初始化创建一个流水线(列表）。
        """
        self.line = []  # 流水线
        self.count = 0  # block id
        self.current_num = 0
        self.__length = None
        self.ready = False

    @property
    def length(self):
        self.__length = len(self.line)
        return self.__length

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_num < len(self.line):
            ret = self.line[self.current_num]
            self.current_num += 1
            return ret
        else:
            raise StopIteration

    def add(self, func, args=(), kwargs=None, inherit=True):
        """
        向流水线中添加块。
        :param func: 加工方法
        :param args: *arg参数
        :param kwargs: **kwargs参数
        :param inherit: 是否继承，
        :return: None
        特别说明：参数inherit优先级大于args, kwargs即当inherit=Ture时，函数会传入上一个函数的返回值，此时指定的args,kwargs均无效。
        """
        if kwargs is None:
            kwargs = {}

        block = StreamBlock(func, inherit, args, kwargs)
        self.line.append(
            {
                'id': self.count,
                'status': block.status,
                'block': block
            }
        )
        self.count += 1

    def append(self, blocks: list):
        """
        批量提交block
        :param blocks: block列表
        :return: None
        """
        if isinstance(blocks, list):
            for b in blocks:
                block = None
                if isinstance(b, tuple):
                    block = StreamBlock(*b)
                elif isinstance(b, dict):
                    block = StreamBlock(**b)
                self.line.append(
                    {
                        'id': self.count,
                        'status': block.status,
                        'block': block
                    }
                )
                self.count += 1

    def start(self, target=None):
        """
        开启流水线加工
        :param target: 原材料
        :return: 成品(数据) | Error
        """
        value = self.line[0]['block'].start(target)
        if self.line[0]['status']['status'] == 'error':
            return 'Error'

        for stuff in self.line[1:]:
            value = stuff['block'].start(value)
            if stuff['status']['status'] == 'error':
                return 'Error'
        self.ready = True
        return value

    def __str__(self):
        return str(self.line)



import time


class StreamBlock:
    """
        块, 用于对一个未知的函数/方法座一层封装，以便于维护、管理改函数/方法的状态
    """
    def __init__(self, func, mode=True, args=None, kwargs=None):
        """
        初始化
        :param func: 自定义方法
        :param mode: 是否继承上一个方法的返回值
        :param args: args
        :param kwargs: kwargs
        """
        if kwargs is None:
            kwargs = {}

        if args is None:
            args = ()

        self.func = func
        self.mode = mode
        self.status = {'status': '', 'result': '', 'cost': '', 'mode': self.mode, 'param': ''}
        self.args = args
        self.kwargs = kwargs
        if args or kwargs:
            self.mode = False

    def start(self, target=None):
        """
        执行内部方法
        :param target:
        :return: Any
        """
        start_time = time.perf_counter()
        self.status['status'] = 'running'
        try:
            # 情况1 继承，没有传入参数, target为None, 适用于不传参的方法。
            if self.mode and self.args == () and self.kwargs == {} and target is None:
                value = self.func()

            # 情况2 继承 使用target
            elif self.mode:
                value = self.func(target)
                self.status['param'] = str(target)

            # 情况3，不继承，传入自定义参数， 不使用target
            else:
                value = self.func(*self.args, **self.kwargs)
                self.status['param'] = str(f'args:{str(self.args)}, kwargs:{str(self.kwargs)}')

            self.status['cost'] = '%sms' % ((time.perf_counter() - start_time) * 1000)
            self.status['result'] = value
            self.status['status'] = 'completed'
            return value
        except Exception as e:
            self.status['status'] = 'error'
            self.status['result'] = str(e)
        return None



from work_stream.line import StreamLine


class StreamBranch:
    """
        流分支，用于判断用哪条流水线。
    """
    def __init__(self, expression, branch1: list[StreamLine, ...], branch2: list[StreamLine, ...]):
        """
        创建分支选择
        :param expression: 分支表达式
        :param branch1: 分支1
        :param branch2: 分支2
        """
        self.branch1 = branch1
        self.branch2 = branch2
        self.expression = expression

    def start(self, value) -> list[StreamLine, ...]:
        """
        分支选择
        :param value: 上一个函数的返回值
        :return: 分支
        """
        res = (self.branch1, self.branch2)[eval(self.expression)]
        return res



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
      name='work_stream',
      version='0.0.2',
      author='fengqing',
      author_email='fq20210602@gmail.com',
      url="https://github.com/fengqing2021/work-stream",
      description='这是一个运维工作流模块',
      packages=find_packages(),
      long_description=long_description,
      long_description_content_type="text/markdown",
      zip_safe=True,
      python_requires=">=3.6",
)



from work_stream.branch import StreamBranch
from work_stream.line import StreamLine
from copy import deepcopy


class StreamContainer:
    """
        流水线容器， 用于给多个流水线分组/打包/编排
    """

    def __init__(self):
        """
        初始化容器
        """
        self.container = {}
        self.count = 0

    def create(self, name='default'):
        """
        创建容器，初始化容器状态
        :param name: 容器名称
        :return: self
        """
        self.container[name] = {}
        self.container[name]['lines'] = {}
        self.container[name]['status'] = {}
        self.container[name]['line_group'] = []
        return self

    def layout(self, layout_line: StreamLine | StreamBranch | list[StreamLine | StreamBranch, ...],
               container_name: str = 'default'):
        """
        编排流水线
        :param layout_line: 流水线, 分支
        :param container_name: 容器名称
        :return: None
        """
        if isinstance(layout_line, list):
            for i in layout_line:
                if not isinstance(i, StreamBranch | StreamLine):
                    raise TypeError(f'该"{i}"参数不为StreamBranch类型或StreamLine类型, 请传入一个正确的参数')
            self.container[container_name]['line_group'] += layout_line
        elif isinstance(layout_line, StreamLine) or isinstance(layout_line, StreamBranch):
            self.container[container_name]['line_group'].append(layout_line)

    def start(self, target=None, container_name='default'):
        """
        启动容器
        :param target:
        :param container_name: 容器名
        :return: Any
        """
        group = self.container[container_name]['line_group']
        assert not isinstance(group[0], StreamBranch), '第一个节点不能为分支'
        self.container[container_name]['status'][self.count] = deepcopy(group[0].line)
        value = group[0].start(target)
        if value == 'Error':
            return None

        for g in range(len(group[1:])):
            if isinstance(group[g + 1], StreamLine):
                self.count += 1
                self.container[container_name]['status'][self.count] = deepcopy(group[g + 1].line)
                value = group[g + 1].start(value)
                if value == 'Error':
                    return None

            elif isinstance(group[g + 1], StreamBranch):
                lines = group[g + 1].start(value)
                for i in lines:
                    self.count += 1
                    self.container[container_name]['status'][self.count] = deepcopy(i.line)
                    value = i.start(value)
                    if value == 'Error':
                        return None
        return value
